@@ -2,19 +2,22 @@ import fim.user.topology as topo
 import fim.user.node as node
 import fim.user.link as link
 import fim.user.component as comp
-import fim.user.switch_fabric as sf
+import fim.user.network_service as ns
 import fim.slivers.network_node as nnode
 import fim.slivers.network_link as nlink
+import fim.slivers.network_service as nss
 import fim.slivers.attached_components as atcomp
 import fim.slivers.component_catalog as cata
 import fim.slivers.interface_info as iinfo
 import fim.slivers.capacities_labels as caplab
 import fim.slivers.delegations as dlg
+import fim.graph.abc_property_graph as abcpg
 
 # push some definitions up to simplify import management
 
 ExperimentTopology = topo.ExperimentTopology
 SubstrateTopology = topo.SubstrateTopology
+AdvertisedTopology = topo.AdvertizedTopology
 Node = node.Node
 Component = comp.Component
 NodeType = nnode.NodeType
@@ -23,11 +26,16 @@ ComponentCatalog = cata.ComponentCatalog
 InterfaceType = iinfo.InterfaceType
 Capacities = caplab.Capacities
 Labels = caplab.Labels
-Layer = sf.SFLayer
+ReservationInfo = caplab.ReservationInfo
+Layer = nss.NSLayer
 Link = link.Link
 LinkType = nlink.LinkType
+ServiceType = ns.ServiceType
+NetworkService = ns.NetworkService
 TopologyDetail = topo.TopologyDetail
 Delegation = dlg.Delegation
+Delegations = dlg.Delegations
 DelegationType = dlg.DelegationType
 Pool = dlg.Pool
-ARMPools = dlg.ARMPools
+Pools = dlg.Pools
+GraphFormat = abcpg.GraphFormat
