@@ -24,7 +24,7 @@
 #
 # Author: Ilya Baldin (ibaldin@renci.org)
 
-from abc import ABC, abstractmethod
+from abc import ABC
 from typing import List, Dict, Tuple, Any
 import enum
 
@@ -38,18 +38,20 @@ from fim.view_only_dict import ViewOnlyDict
 from .model_element import ModelElement
 from ..slivers.network_node import NodeType
 from ..slivers.network_link import LinkType
-from ..slivers.switch_fabric import SFLayer
+from ..slivers.network_service import NSLayer, ServiceType
 from ..graph.slices.abc_asm import ABCASMPropertyGraph
 from ..graph.slices.networkx_asm import NetworkxASM
-from ..graph.abc_property_graph import ABCPropertyGraph
-from ..graph.resources.networkx_adm import NetworkXADMGraph, NetworkXGraphImporter
-from ..slivers.delegations import Delegation, ARMDelegations, ARMPools, DelegationType
+from ..graph.abc_property_graph import ABCPropertyGraph, GraphFormat
+from ..graph.resources.networkx_arm import NetworkXARMGraph, NetworkXGraphImporter
+from ..slivers.delegations import Delegation, Delegations, Pools, DelegationType, DelegationFormat
 from fim.graph.resources.networkx_abqm import NetworkXAggregateBQM, NetworkXABQMFactory
 from fim.slivers.capacities_labels import Capacities, CapacityTuple
 
 from .model_element import ElementType
 from .node import Node
+from .component import Component
 from .composite_node import CompositeNode
+from .network_service import NetworkService
 from .interface import Interface
 from .link import Link
 
@@ -76,6 +78,79 @@ class Topology(ABC):
         if graph_file is not None or graph_string is not None:
             self.load(file_name=graph_file, graph_string=graph_string)
 
+    def get_parent_element(self, e: ModelElement) -> ModelElement or None:
+        """
+        Instantiate a ModelElement parent of this element or return None
+        for Links, Nodes and sometimes NetworkServices
+        :param e:
+        :return:
+        """
+        assert e is not None
+        if isinstance(e, Node) or isinstance(e, Link):
+            # nodes or links don't have parents
+            return None
+        # components have nodes as parents
+        if isinstance(e, Component):
+            node_name, node_id = self.graph_model.get_parent(node_id=e.node_id,
+                                                             rel=ABCPropertyGraph.REL_HAS,
+                                                             parent=ABCPropertyGraph.CLASS_NetworkNode)
+            if node_id is None:
+                raise RuntimeError(f'Component {e} has no parent')
+            return Node(name=node_name, node_id=node_id, topo=self)
+        # NSs may have nodes or components as parents or be by themselves
+        if isinstance(e, NetworkService):
+            node_name, node_id = self.graph_model.get_parent(node_id=e.node_id,
+                                                             rel=ABCPropertyGraph.REL_HAS,
+                                                             parent=ABCPropertyGraph.CLASS_Component)
+            if node_id is not None:
+                return Component(name=node_name, node_id=node_id, topo=self)
+
+            node_name, node_id = self.graph_model.get_parent(node_id=e.node_id,
+                                                             rel=ABCPropertyGraph.REL_HAS,
+                                                             parent=ABCPropertyGraph.CLASS_NetworkNode)
+            if node_id is not None:
+                return Node(name=node_name, node_id=node_id, topo=self)
+            else:
+                return None
+        # interfaces have NSs as parents
+        if isinstance(e, Interface):
+            node_name, node_id = self.graph_model.get_parent(node_id=e.node_id,
+                                                             rel=ABCPropertyGraph.REL_CONNECTS,
+                                                             parent=ABCPropertyGraph.CLASS_NetworkService)
+            if node_id is None:
+                raise RuntimeError(f'Interface {e} has no parent')
+            return NetworkService(name=node_name, node_id=node_id, topo=self)
+        raise RuntimeError(f'Unable to determine parent of element {e}')
+
+    def get_owner_node(self, e: ModelElement) -> ModelElement or None:
+        """
+        Get an owner Node element for Component, NetworkService or Interface. For Link and Node
+        return None
+        :param e:
+        :return:
+        """
+        assert e is not None
+        if isinstance(e, Node) or isinstance(e, Link):
+            # nodes or links don't have parents
+            return None
+        if isinstance(e, Component):
+            return self.get_parent_element(e)
+        if isinstance(e, NetworkService):
+            # either it is directly owned by a switch or via a component or by itself
+            pe = self.get_parent_element(e)
+            if pe is None:
+                return None
+            if isinstance(pe, Node):
+                return pe
+            else:
+                return self.get_parent_element(pe)
+        if isinstance(e, Interface):
+            # figure out the owner of the parent network service
+            ns = self.get_parent_element(e)
+            if ns is not None:
+                return self.get_owner_node(ns)
+        return RuntimeError(f'Unable to determine owner node of element {e}')
+
     def add_node(self, *, name: str, node_id: str = None, site: str, ntype: NodeType = NodeType.VM,
                  **kwargs) -> Node:
         """
@@ -107,26 +182,26 @@ class Topology(ABC):
         :return:
         """
         assert name is not None
-        self.graph_model.remove_network_node_with_components_sfs_cps_and_links(
+        self.graph_model.remove_network_node_with_components_nss_cps_and_links(
             node_id=self.__get_node_by_name(name=name).node_id)
 
-    def add_link(self, *, name: str, node_id: str = None, ltype: LinkType = None,
-                 interfaces: List[Interface], layer: SFLayer = None, technology: str = None,
+    def add_link(self, *, name: str, node_id: str = None, ltype: LinkType,
+                 interfaces: List[Interface], technology: str = None,
                  **kwargs) -> Link:
         """
-        Add link between listed interfaces with specified parameters
+        Add link between listed interfaces with specified parameters (for experiment
+        topologies you want add_network_service instead)
         :param name:
         :param node_id:
         :param ltype:
         :param interfaces:
-        :param layer:
         :param technology:
         :param kwargs:
         :return:
         """
         # add link to graph
-        link = Link(name=name, node_id=node_id, ltype=ltype, interfaces=interfaces, **kwargs,
-                    etype=ElementType.NEW, topo=self, layer=layer, technology=technology)
+        link = Link(name=name, node_id=node_id, ltype=ltype, interfaces=interfaces,
+                    etype=ElementType.NEW, topo=self, technology=technology, **kwargs)
         return link
 
     def remove_link(self, name: str):
@@ -138,6 +213,31 @@ class Topology(ABC):
         assert name is not None
         self.graph_model.remove_network_link(node_id=self.__get_link_by_name(name=name).node_id)
 
+    def add_network_service(self, *, name: str, node_id: str = None, nstype: ServiceType,
+                            interfaces: List[Interface], technology: str = None, **kwargs) -> NetworkService:
+        """
+        Add a network service to a topology. Interfaces can be specified upfront or added/removed later
+        :param name:
+        :param node_id:
+        :param nstype:
+        :param interfaces:
+        :param technology:
+        :param kwargs:
+        :return:
+        """
+        ns = NetworkService(name=name, node_id=node_id, topo=self, nstype=nstype,
+                            etype=ElementType.NEW, interfaces=interfaces, technology=technology)
+        return ns
+
+    def remove_network_service(self, name: str):
+        """
+        Remove a network service and associated service interfaces
+        :param name:
+        :return:
+        """
+        assert name is not None
+        self.graph_model.remove_ns_with_cps_and_links(node_id=self.__get_ns_by_name(name=name).node_id)
+
     def __get_node_by_name(self, name: str) -> Node:
         """
         Find node by its name, return Node object
@@ -160,6 +260,17 @@ class Topology(ABC):
                                                      label=ABCPropertyGraph.CLASS_Link)
         return Link(name=name, node_id=node_id, topo=self)
 
+    def __get_ns_by_name(self, name: str) -> NetworkService:
+        """
+        Find a network service by its name, return NetworkService object
+        :param name:
+        :return:
+        """
+        assert name is not None
+        node_id = self.graph_model.find_node_by_name(node_name=name,
+                                                     label=ABCPropertyGraph.CLASS_NetworkService)
+        return NetworkService(name=name, node_id=node_id, topo=self)
+
     def __get_node_by_id(self, node_id: str) -> Node:
         """
         Get node by its node_id, return Node object
@@ -182,6 +293,17 @@ class Topology(ABC):
         assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
         return Link(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id, topo=self)
 
+    def __get_ns_by_id(self, node_id: str) -> NetworkService:
+        """
+        Get network service by its node id, return NetworkService object
+        :param node_id:
+        :return:
+        """
+        assert node_id is not None
+        _, node_props = self.graph_model.get_node_properties(node_id=node_id)
+        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
+        return NetworkService(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id, topo=self)
+
     def __list_nodes(self) -> ViewOnlyDict:
         """
         List all NetworkNodes in the topology as a dictionary
@@ -196,6 +318,20 @@ class Topology(ABC):
             ret[n.name] = n
         return ViewOnlyDict(ret)
 
+    def __list_network_services(self) -> ViewOnlyDict:
+        """
+        List all NetworkServices in the topology as a dictionary organized by name.
+        Modifying the dictionary will not affect the underlying model, but modifying
+        NetworkServices in the dictionary will.
+        :return:
+        """
+        node_id_list = self.graph_model.get_all_network_service_nodes()
+        ret = dict()
+        for nid in node_id_list:
+            n = self.__get_ns_by_id(nid)
+            ret[n.name] = n
+        return ViewOnlyDict(ret)
+
     def __list_links(self) -> ViewOnlyDict:
         """
         List all Links in the topology as a dictionary organized by Link name.
@@ -231,18 +367,21 @@ class Topology(ABC):
             return self.__list_nodes()
         if item == 'links':
             return self.__list_links()
+        if item == 'network_services':
+            return self.__list_network_services()
         if item == 'interface_list':
             return self.__list_of_interfaces()
         raise RuntimeError(f'Attribute {item} not available')
 
-    def serialize(self, file_name=None) -> str or None:
+    def serialize(self, file_name: str = None, fmt: GraphFormat = GraphFormat.GRAPHML) -> str or None:
         """
         Serialize to string or to file, depending on whether file_name
         is provided.
         :param file_name:
+        :param fmt: one of GraphFormat possible values
         :return string containing GraphML if file_name is None:
         """
-        graph_string = self.graph_model.serialize_graph()
+        graph_string = self.graph_model.serialize_graph(format=fmt)
         if file_name is None:
             return graph_string
         else:
@@ -307,23 +446,33 @@ class Topology(ABC):
         elif topo_detail == TopologyDetail.Derived:
             # collect all network nodes and links, draw edges between them
             network_nodes = self.graph_model.get_all_network_nodes()
+            service_nodes = self.graph_model.get_all_network_service_nodes()
             links = self.graph_model.get_all_network_links()
 
             derived_graph = nx.Graph()
             for n in network_nodes:
                 _, props = self.graph_model.get_node_properties(node_id=n)
                 derived_graph.add_node(props[ABCPropertyGraph.PROP_NAME])
-            for l in links:
-                _, props = self.graph_model.get_node_properties(node_id=l)
+            for ns in service_nodes:
+                _, props = self.graph_model.get_node_properties(node_id=ns)
                 derived_graph.add_node(props[ABCPropertyGraph.PROP_NAME])
 
-            for n in self.nodes.values():
-                node_ints = n.interfaces.values()
-                for nint in node_ints:
-                    for l in self.links.values():
-                        # this works because of custom ModelElement.__eq__()
-                        if nint in l.interface_list:
-                            derived_graph.add_edge(n.name, l.name)
+            # link interfaces of NSs and NSs to nodes
+            for ns in self.network_services.values():
+                for nint in ns.interfaces.values():
+                    peer_int_id = self.graph_model.find_peer_connection_point(node_id=nint.node_id)
+                    if peer_int_id is None:
+                        continue
+                    _, peer_props = self.graph_model.get_node_properties(node_id=peer_int_id)
+                    peer_int = Interface(node_id=peer_int_id, name=peer_props[ABCPropertyGraph.PROP_NAME],
+                                         topo=self)
+                    peer_int_parent = self.get_parent_element(peer_int)
+                    if peer_int_parent is None:
+                        continue
+                    derived_graph.add_edge(ns.name, peer_int_parent.name)
+                for n in self.nodes.values():
+                    if self.get_owner_node(ns) == n:
+                        derived_graph.add_edge(ns.name, n.name)
 
             pos = layout(derived_graph)
             nx.draw_networkx(derived_graph, pos=pos)
@@ -334,6 +483,13 @@ class Topology(ABC):
         else:
             raise RuntimeError("This level of detail not yet implemented")
 
+    @staticmethod
+    def __print_caplabs__(caps) -> str:
+        if caps is None:
+            return '{ }'
+        else:
+            return str(caps)
+
     def __str__(self):
         """
         Print topology in tabulated form - network nodes, their components, interfaces, then print links
@@ -342,18 +498,18 @@ class Topology(ABC):
         """
         lines = list()
         for n in self.nodes.values():
-            lines.append(n.name + "[" + str(n.get_property("type")) + "]: " + str(n.get_property("capacities")))
+            lines.append(n.name + "[" + str(n.get_property("type")) + "]:  " +
+                         self.__print_caplabs__(n.get_property("capacities")))
             for i in n.direct_interfaces.values():
                 lines.append("\t\t" + i.name + ": " + str(i.get_property("type")) + " " +
-                             str(i.get_property("capacities")) + " " +
-                             str(i.get_property("labels")))
+                             self.__print_caplabs__(i.get_property("capacities")) + " " +
+                             self.__print_caplabs__(i.get_property("labels")))
             for c in n.components.values():
                 lines.append("\t" + c.name + ": " + " " + str(c.get_property("type")) + " " +
                              c.get_property("model"))
-                print("Interfaces:")
                 for i in c.interfaces.values():
                     lines.append("\t\t" + i.name + ": " + str(i.get_property("type")) + " " +
-                                 str(i.get_property("capacities")))
+                                 self.__print_caplabs__(i.get_property("capacities")))
         lines.append("Links:")
         for l in self.links.values():
             interface_names = [iff.name for iff in l.interface_list]
@@ -399,45 +555,48 @@ class SubstrateTopology(Topology):
     def __init__(self, graph_file: str = None, graph_string: str = None, logger=None):
         super().__init__(graph_file=graph_file, graph_string=graph_string, logger=logger)
 
-    def as_adm(self):
+    def as_arm(self):
         """
-        Return model as ADM graph built on top of internal model. Model is not cloned or copied,
-        rather recast into ADM, so any changes to the model propagate back to the topology.
+        Return model as ARM graph built on top of internal model. Model is not cloned or copied,
+        rather recast into ARM, so any changes to the model propagate back to the topology.
         :return:
         """
-        return NetworkXADMGraph(graph_id=self.graph_model.graph_id,
-                                importer=self.graph_model.importer,
+        return NetworkXARMGraph(graph=self.graph_model,
                                 logger=self.graph_model.log)
 
     @staticmethod
-    def __copy_to_delegation(e: ModelElement, atype: DelegationType,
-                             delegation_id: str) -> Delegation or None:
+    def __copy_to_delegations(e: ModelElement, atype: DelegationType,
+                              delegation_id: str) -> Delegations or None:
         """
-        From a model element (node, components, switch_fabric, interface), copy
-        capacities or labels to delegation and return delegation)
+        From a model element (node, component, switch_fabric, interface), copy
+        capacities or labels to a Delegations and return Delegations or none if
+        no capacities or labels specified. Return None if element is a stitch node.
         :param e:
         :return:
         """
         assert e is not None
         assert delegation_id is not None
-        ret = Delegation(atype=atype, defined_on=e.node_id, delegation_id=delegation_id)
+        if e.get_property("stitch_node"):
+            return None
         if atype == DelegationType.CAPACITY:
             caps_or_labels = e.get_property(pname='capacities')
-            if caps_or_labels is not None:
-                ret.set_details_from_capacities(caps_or_labels)
-                return ret
         else:
             caps_or_labels = e.get_property(pname='labels')
-            if caps_or_labels is not None:
-                ret.set_details_from_labels(caps_or_labels)
-                return ret
+        if caps_or_labels is not None:
+            d = Delegation(atype=atype, delegation_id=delegation_id, aformat=DelegationFormat.SinglePool)
+            d.set_details(caps_or_labels)
+            # encapsulate in Delegations - this is a single delegation case, no others will be added
+            ret = Delegations(atype=atype)
+            ret.add_delegations(d)
+            return ret
         return None
 
-    def single_delegation(self, *, delegation_id: str, label_pools: ARMPools, capacity_pools: ARMPools):
+    def single_delegation(self, *, delegation_id: str, label_pools: Pools, capacity_pools: Pools):
         """
         For simple cases when there is one delegation that delegates everything. Be sure to have
         capacities and labels specified on individual elements where needed - they get copied
-        into delegations. Pools must be specified externally.
+        into delegations. Pools must be specified externally. There is no way to specify on
+        the same node a single element delegation and a pool definition or reference.
         :param delegation_id:
         :param label_pools:
         :param capacity_pools:
@@ -446,42 +605,42 @@ class SubstrateTopology(Topology):
         assert label_pools.pool_type == DelegationType.LABEL
         assert capacity_pools.pool_type == DelegationType.CAPACITY
 
-        delegations = {DelegationType.CAPACITY: ARMDelegations(DelegationType.CAPACITY),
-                       DelegationType.LABEL: ARMDelegations(DelegationType.LABEL)}
+        delegations_dicts = {DelegationType.CAPACITY: dict(),
+                             DelegationType.LABEL: dict()}
         pool_dict = {DelegationType.CAPACITY: capacity_pools,
                      DelegationType.LABEL: label_pools}
         for t in DelegationType:
             for n in self.nodes.values():
-                # delegate every node and its component by copying their capacities and labels
-                # if present
-                delegation = self.__copy_to_delegation(e=n, atype=t,
-                                                       delegation_id=delegation_id)
-                if delegation is not None:
-                    delegations[t].add_delegation(delegation)
+                # delegate every NetworkNode and its component by copying their capacities and labels
+                # if present. Each delegations object is intended for one node.
+                delegations = self.__copy_to_delegations(e=n, atype=t,
+                                                         delegation_id=delegation_id)
+                if delegations is not None:
+                    delegations_dicts[t][n.node_id] = delegations
                 # components
                 for c in n.components.values():
-                    delegation = self.__copy_to_delegation(e=c, atype=t,
-                                                           delegation_id=delegation_id)
-                    if delegation is not None:
-                        delegations[t].add_delegation(delegation)
+                    delegations = self.__copy_to_delegations(e=c, atype=t,
+                                                             delegation_id=delegation_id)
+                    if delegations is not None:
+                        delegations_dicts[t][c.node_id] = delegations
                     for i in c.interfaces.values():
-                        delegation = self.__copy_to_delegation(e=i, atype=t,
-                                                               delegation_id=delegation_id)
-                        if delegation is not None:
-                            delegations[t].add_delegation(delegation)
-                # switchfabrics (for eg switches)
-                for sf in n.switch_fabrics.values():
-                    delegation = self.__copy_to_delegation(e=sf, atype=t,
-                                                           delegation_id=delegation_id)
-                    if delegation is not None:
-                        delegations[t].add_delegation(delegation)
+                        delegations = self.__copy_to_delegations(e=i, atype=t,
+                                                                 delegation_id=delegation_id)
+                        if delegations is not None:
+                            delegations_dicts[t][i.node_id] = delegations
+                # network services (for eg switches)
+                for sf in n.network_services.values():
+                    delegations = self.__copy_to_delegations(e=sf, atype=t,
+                                                             delegation_id=delegation_id)
+                    if delegations is not None:
+                        delegations_dicts[t][sf.node_id] = delegations
                     for i in sf.interfaces.values():
-                        delegation = self.__copy_to_delegation(e=i, atype=t,
-                                                               delegation_id=delegation_id)
-                        if delegation is not None:
-                            delegations[t].add_delegation(delegation)
+                        delegations = self.__copy_to_delegations(e=i, atype=t,
+                                                                 delegation_id=delegation_id)
+                        if delegations is not None:
+                            delegations_dicts[t][i.node_id] = delegations
 
-            self.as_adm().annotate_delegations_and_pools(dels=delegations[t],
+            self.as_arm().annotate_delegations_and_pools(dels=delegations_dicts[t],
                                                          pools=pool_dict[t])
 
 
@@ -507,7 +666,7 @@ class AdvertizedTopology(Topology):
         raise RuntimeError('Cannot add node to advertisement')
 
     def add_link(self, *, name: str, node_id: str = None, ltype: LinkType = None,
-                 interfaces: List[Interface], layer: SFLayer = None, technology: str = None,
+                 interfaces: List[Interface], layer: NSLayer = None, technology: str = None,
                  **kwargs) -> Link:
         raise RuntimeError('Cannot add link to advertisement')
 
